# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103
#
#  Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                 by The Finc Authors, http://finc.info
#                 by Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
IMDB.
"""

from siskin.common import FTPMirror
from gluish.intervals import hourly
from siskin.task import DefaultTask
import luigi

class IMDBTask(DefaultTask):
    TAG = 'imdb'

class IMDBSync(IMDBTask):

    indicator = luigi.Parameter(default=hourly(fmt='%s'))

    def requires(self):
        return FTPMirror(host='ftp.fu-berlin.de',
                         username='anonymous',
                         password='',
                         base='/pub/misc/movies/database/')

    def run(self):
        self.input().move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path())
